/* eslint-disable */
/**
 * This method returns a SAX-like HTML parser.
 * This code is copied from the HTML Parser writted by John Resig (ejohn.org).
 * The original code is licenced under the Mozilla Public License.
 *
 * The following modifications were introduced:
 * - Remove HTMLtoXML and HTMLtoDOM methods
 * - Modify RegExp to accept the "$" sign in the tag names and attributes
 * - Change handling of special tags (style, script, code).
 *   The text content is not parsed and returned "as-is".
 */
export function newHtmlParser({

  // If this flag is "true" then the parser handles multiple root tags.
  // Otherwise it stops parsing when the first tag is finished.
  multipleRoots = true,

  // Special Elements (can contain anything)
  specialElements = "code,script,style",

  // Empty Elements - HTML 4.01
  emptyElements = "area,base,basefont,br,col,frame,hr,img,input,isindex,link,meta,param,embed",

  // Block Elements - HTML 4.01
  blockElements = "address,applet,blockquote,button,center,dd,del,dir,div,dl,dt,fieldset,form,frameset,hr,iframe,ins,isindex,li,map,menu,noframes,noscript,object,ol,p,pre,script,table,tbody,td,tfoot,th,thead,tr,ul",

  // Inline Elements - HTML 4.01
  inlineElements = "a,abbr,acronym,applet,b,basefont,bdo,big,br,button,cite,code,del,dfn,em,font,i,iframe,img,input,ins,kbd,label,map,object,q,s,samp,script,select,small,span,strike,strong,sub,sup,textarea,tt,u,var",

  // Elements that you can, intentionally, leave open
  // (and which close themselves)
  closeSelfElements = "colgroup,dd,dt,li,options,p,td,tfoot,th,thead,tr",

  // Attributes that have their values filled in disabled="disabled"
  fillAttributes = "checked,compact,declare,defer,disabled,ismap,multiple,nohref,noresize,noshade,nowrap,readonly,selected",

} = { }) {
  /*
  * HTML Parser By John Resig (ejohn.org)
  * Original code by Erik Arvidsson, Mozilla Public License
  * http://erik.eae.net/simplehtmlparser/simplehtmlparser.js
  *
  * // Use like so:
  * HTMLParser(htmlString, {
  *     start: function(tag, attrs, unary) {},
  *     end: function(tag) {},
  *     chars: function(text) {},
  *     comment: function(text) {}
  * });
  *
  * // or to get an XML string:
  * HTMLtoXML(htmlString);
  *
  * // or to get an XML DOM Document
  * HTMLtoDOM(htmlString);
  *
  * // or to inject into an existing document/DOM node
  * HTMLtoDOM(htmlString, document);
  * HTMLtoDOM(htmlString, document.body);
  *
  */
 
  // Regular Expressions for parsing tags and attributes
  var startTag = /^<([A-Za-z_\{\}$][-A-Za-z0-9_:\.\{\}$]*?)((?:\s+[\w-A-Za-z0-9_:\.\{\}$]+(?:\s*=\s*(?:(?:"[^"]*")|(?:'[^']*')|[^\/>\s]*))?)*)\s*(\/?)>/,
    endTag = /^<\/([A-Za-z_\{\}$][-A-Za-z0-9_:\.\{\}$]*)[^>]*>/,
    attr = /([A-Za-z_\{\}$][-A-Za-z0-9_:\.\{\}$]*)(?:\s*=\s*(?:(?:"((?:\\.|[^"])*)")|(?:'((?:\\.|[^'])*)')|([^\/>\s]*)))?/g;

  // Empty Elements - HTML 4.01
  var empty = makeMap(emptyElements);

  // Block Elements - HTML 4.01
  var block = makeMap(blockElements);

  // Inline Elements - HTML 4.01
  var inline = makeMap(inlineElements);

  // Elements that you can, intentionally, leave open
  // (and which close themselves)
  var closeSelf = makeMap(closeSelfElements);

  // Attributes that have their values filled in disabled="disabled"
  var fillAttrs = makeMap(fillAttributes);

  // Special Elements (can contain anything)
  var special = makeMap(specialElements);

  var HTMLParser = function (html, handler) {
    var index, chars, match, stack = [], last = html;
    stack.last = function () {
      return this[this.length - 1];
    };

    const source = html;

    while (html) {
      chars = true;

      // Make sure we're not in a script or style element
      if (!stack.last() || !special[stack.last()]) {

        // XML declaration
        if (html.indexOf("<?xml") == 0) {
          index = html.indexOf("?>");
          if (index >= 0) {
            html = html.substring(index + 2);
            chars = false;
          }

        } else if (html.indexOf("<!--") == 0) { // Comment
          index = html.indexOf("-->");

          if (index >= 0) {
            if (handler.comment)
              handler.comment(html.substring(4, index));
            html = html.substring(index + 3);
            chars = false;
          }

          // end tag
        } else if (html.indexOf("</") == 0) {
          match = html.match(endTag);

          if (match) {
            html = html.substring(match[0].length);
            match[0].replace(endTag, parseEndTag);
            chars = false;
          }

          // start tag
        } else if (html.indexOf("<") == 0) {
          match = html.match(startTag);

          if (match) {
            html = html.substring(match[0].length);
            match[0].replace(startTag, parseStartTag);
            chars = false;
          }
        }

        if (chars) {
          index = html.indexOf("<");

          var text = index < 0 ? html : html.substring(0, index);
          html = index < 0 ? "" : html.substring(index);

          if (handler.chars)
            handler.chars(text);
        }

      } else {
        //         html = html.replace(new RegExp("(.*)<\/" + stack.last() + "[^>]*>"), function(all, text){
        //           text = text.replace(/<!--(.*?)-->/g, "$1")
        //             .replace(/<!\[CDATA\[(.*?)]]>/g, "$1");

        //           if ( handler.chars )
        //             handler.chars( text );

        //           return "";
        //         });

        var idx = 0;
        var lastPos = -1;
        html.replace(new RegExp("<\/" + stack.last() + "[^>]*>", 'gim'), function (match, pos) {
          if (lastPos >= 0) return;
          idx += pos;
          lastPos = idx + match.length;
        });
        if (lastPos > 0) {
          var text = html.substring(0, idx);
          html = html.substring(lastPos);
          text = text.replace(/<!--(.*?)-->/g, "$1")
            .replace(/<!\[CDATA\[(.*?)]]>/g, "$1");
          if (handler.chars)
            handler.chars(text);
        }
        parseEndTag("", stack.last());
      }

      if (html == last) {
        if (!multipleRoots) break; 
        else throw new Error("Parse Error: " + html);
      }
      
      last = html;

      if (!multipleRoots && !stack.length) {
        break;
      }
  
    }

    // Clean up any remaining tags
    parseEndTag();

    if (handler.finish) handler.finish({
      source,
      begin: 0,
      end: source.length - (last || '').length
    })    

    function parseStartTag(tag, tagName, rest, unary) {
      // tagName = tagName.toLowerCase();

      if (block[tagName]) {
        while (stack.last() && inline[stack.last()]) {
          parseEndTag("", stack.last());
        }
      }

      if (closeSelf[tagName] && stack.last() == tagName) {
        parseEndTag("", tagName);
      }

      unary = empty[tagName] || !!unary;
      if (!unary)
        stack.push(tagName);
      if (handler.start) {
        const attrs = [];
        rest.replace(attr, function (match, name) {
          var value = arguments[2] ||
            arguments[3] ||
            arguments[4] ||
            (fillAttrs[name] ? name : undefined);
          attrs.push({ name: name, value: value, });
        });
        handler.start(tagName, attrs, unary);
      }
    }

    function parseEndTag(tag, tagName) {
      // If no tag name is provided, clean shop
      if (!tagName)
        var pos = 0;

      // Find the closest opened tag of the same type
      else
        for (var pos = stack.length - 1; pos >= 0; pos--)
          if (stack[pos] == tagName)
            break;

      if (pos >= 0) {
        // Close all the open elements, up the stack
        for (var i = stack.length - 1; i >= pos; i--)
          if (handler.end)
            handler.end(stack[i]);

        // Remove the open elements from the stack
        stack.length = pos;
      }
    }
  };

  function makeMap(str) {
    var obj = { }, items = str.split(",");
    for (var i = 0; i < items.length; i++)
      obj[items[i]] = true;
    return obj;
  }

  // ==========================================================================

  return (text, handler) => {
    const stack = [];
    HTMLParser(text, {
      finish : handler.finish,
      start: (name, attributes, unary) => {
        const attrs = { };
        attributes.forEach(attr => {
          attrs[attr.name] = attr.value;
        });
        handler.beginTag(name, attrs);
        stack.push([name, attrs]);
        if (unary) { handler.endTag(...stack.pop()); }
      },
      end: () => {
        handler.endTag(...stack.pop());
      },
      chars: (str) => {
        handler.onText(str);
      },
      comment: () => {
      }
    });
    return handler.result;
  }
}
